@@ -9,6 +9,9 @@ pub use progress::Progress;
 use osmpbf_parser::{Parser,Scan,ScanTable,element};
 mod par_scan;
 use par_scan::parallel_scan;
+use eyros::{Point,Value};
+use std::collections::HashMap;
+use async_std::prelude::*;
 
 pub const BACKREF_PREFIX: u8 = 1;
 pub const REF_PREFIX: u8 = 2;
@@ -33,7 +36,7 @@ pub struct IngestOptions {
   pub ingest_node: bool,
   pub ingest_way: bool,
   pub ingest_relation: bool,
-  pub optimize: usize,
+  pub optimize: Option<(usize,usize)>,
 }
 
 impl Default for IngestOptions {
@@ -45,7 +48,7 @@ impl Default for IngestOptions {
       ingest_node: true,
       ingest_way: true,
       ingest_relation: true,
-      optimize: 0,
+      optimize: None,
     }
   }
 }
@@ -68,8 +71,9 @@ impl Ingest {
         Parser::new(Box::new(h))
       }).collect::<Vec<_>>();
       let file_size = std::fs::File::open(pbf_file).unwrap().metadata().unwrap().len();
-      parallel_scan(parsers, 0, file_size).await.unwrap()
+      parallel_scan(self.progress.clone(), parsers, 0, file_size).await.unwrap()
     };
+    self.progress.write().await.add("scan", 0);
     self.progress.write().await.end("scan");
     scan_table
   }
@@ -399,12 +403,111 @@ impl Ingest {
     self.progress.write().await.end("ingest");
   }
 
-  pub async fn optimize(&mut self, mut db: EDB, rebuild_depth: usize) -> Result<(),Error> {
+  pub async fn optimize(
+    &mut self, mut in_db: EDB, mut out_db: EDB, xy_divs: (usize,usize)
+  ) -> Result<(),Error> {
     self.progress.write().await.start("optimize");
-    if rebuild_depth > 0 {
-      db.optimize(rebuild_depth).await?;
-      db.sync().await?;
+    let (x_divs,y_divs) = xy_divs;
+    let ibox = ((f32::INFINITY,f32::INFINITY),(f32::NEG_INFINITY,f32::NEG_INFINITY));
+    let db_bounds = in_db.meta.roots.iter()
+      .filter_map(|r| r.as_ref().map(|tr| tr.bounds.to_bounds().unwrap()))
+      .fold(ibox,|bbox,b| {
+        (
+          ((bbox.0).0.min((b.0).0),(bbox.0).1.min((b.0).1)),
+          ((bbox.1).0.max((b.1).0),(bbox.1).1.max((b.1).1)),
+        )
+      });
+    let db_span = (
+      (db_bounds.1).0 - (db_bounds.0).0,
+      (db_bounds.1).1 - (db_bounds.0).1,
+    );
+    let mut skip = HashMap::new();
+    let mut batch = vec![];
+    let mut sync_count = 0;
+    for iy in 0..y_divs {
+      for ix in 0..x_divs {
+        let bbox = (
+          (
+            (ix as f32)/(x_divs as f32) * db_span.0 + (db_bounds.0).0,
+            (iy as f32)/(y_divs as f32) * db_span.1 + (db_bounds.0).1,
+          ),
+          (
+            ((ix+1) as f32)/(x_divs as f32) * db_span.0 + (db_bounds.0).0,
+            ((iy+1) as f32)/(y_divs as f32) * db_span.1 + (db_bounds.0).1,
+          )
+        );
+        let mut stream = in_db.query(&bbox).await?;
+        let mut count = 0;
+        while let Some(r) = stream.next().await {
+          let (p,v) = r?;
+          if v.is_empty() { continue }
+          let id = v.get_id();
+          if skip.contains_key(&id) {
+            let n = {
+              let n = skip.get_mut(&id).unwrap();
+              *n -= 1;
+              *n
+            };
+            if n <= 0 { skip.remove(&id); }
+            continue;
+          }
+          count += 1;
+          sync_count += 1;
+          let pbounds = (
+            (
+              *match &p.0 {
+                eyros::Coord::Scalar(x) => x,
+                eyros::Coord::Interval(xmin,_) => xmin,
+              },
+              *match &p.1 {
+                eyros::Coord::Scalar(y) => y,
+                eyros::Coord::Interval(ymin,_) => ymin,
+              },
+            ),
+            (
+              *match &p.0 {
+                eyros::Coord::Scalar(x) => x,
+                eyros::Coord::Interval(_,xmax) => xmax,
+              },
+              *match &p.1 {
+                eyros::Coord::Scalar(y) => y,
+                eyros::Coord::Interval(_,ymax) => ymax,
+              },
+            ),
+          );
+          {
+            let mut nx = 1;
+            let mut ny = 1;
+            if (pbounds.0).0 <= (bbox.0).0 || (pbounds.1).0 >= (bbox.1).0 {
+              nx += (((pbounds.1).0 - (pbounds.0).0) / ((bbox.1).0 - (bbox.0).0)).ceil() as usize;
+            }
+            if (pbounds.0).1 <= (bbox.0).1 || (pbounds.1).1 >= (bbox.1).1 {
+              ny += (((pbounds.1).1 - (pbounds.0).1) / ((bbox.1).1 - (bbox.0).1)).ceil() as usize;
+            }
+            let n = nx*ny-1;
+            if n > 0 {
+              skip.insert(id, n);
+            }
+          }
+          batch.push(eyros::Row::Insert(p,v));
+        }
+        if !batch.is_empty() {
+          out_db.batch(&batch).await?;
+          batch.clear();
+        }
+        if sync_count > 1_000_000 {
+          out_db.sync().await?;
+          sync_count = 0;
+        }
+        if count > 0 {
+          self.progress.write().await.add("optimize", count);
+        }
+      }
     }
+    if sync_count > 0 {
+      out_db.sync().await?;
+    }
+    self.progress.write().await.add("optimize", 0);
     self.progress.write().await.end("optimize");
     Ok(())
   }
